import math


def linear(step, total_steps):
    """ linearly increasing from [0 to 1] """
    return step / max(1, total_steps - 1)


def cosine(step, total_steps):
    """ cosine schedule from [0 to 1] """
    progress = step / max(1, total_steps - 1)
    return 1 - (1 + math.cos(math.pi * progress)) / 2


def inverse_sqrt(step, abs_step):
    """ inverse square root schedule starting from 1 and going towards 0 """
    # https://github.com/facebookresearch/fairseq/blob/main/fairseq/optim/lr_scheduler/inverse_square_root_schedule.py
    decay_factor  = max(1, abs_step - step) ** 0.5
    if step == abs_step:
        cur_step = abs_step + 1
    else:
        cur_step = abs_step
    return cur_step ** -0.5 * decay_factor


from .progress_schedule import ProgressSchedule


class DecreasingProgressSchedule(ProgressSchedule):
    def __init__(self, max_value=1., end_value=0., **kwargs):
        delta = end_value - max_value
        assert delta <= 0.
        super().__init__(start_value=max_value, delta=delta, **kwargs)

    def _get_progress(self, step: int, total_steps: int, abs_step: int):
        raise NotImplementedError



from .base import DecreasingProgressSchedule

from .functional.schedules import linear


class LinearDecreasingSchedule(DecreasingProgressSchedule):
    def _get_progress(self, step: int, total_steps: int, abs_step: int):
        return linear(step, total_steps)



from .base import ScheduleBase


class StepIntervalSchedule(ScheduleBase):
    def __init__(self, start_value, factor, interval):
        super().__init__()
        assert isinstance(interval, (int, float)) and 0. < interval < 1.
        self.start_value = start_value
        self.factor = factor
        self.interval = interval

    def __str__(self):
        return f"{type(self).__name__}(start_value={self.start_value}, factor={self.factor}, interval={self.interval})"

    def _get_value(self, step: int, total_steps: int, abs_step: int = None) -> float:
        progress = step / total_steps
        # round to 10th decimal place to avoid floating point precision errors
        step_idx = int(round(progress / self.interval, 10))
        return self.start_value * self.factor ** step_idx



import kappaschedules.schedules
from .schedules import (
    SequentialPercentSchedule,
    SequentialPercentScheduleConfig,
    SequentialStepSchedule,
    SequentialStepScheduleConfig,
)
from .schedules.base import ScheduleBase
import inspect
from copy import deepcopy

def object_to_schedule(obj, batch_size=None, updates_per_epoch=None, **kwargs) -> ScheduleBase:
    if obj is None:
        return None
    if not isinstance(obj, (list, dict)):
        assert isinstance(obj, ScheduleBase)
        return obj
    obj = deepcopy(obj)

    # implicit sequential schedule
    if isinstance(obj, list):
        # check consistency between step/percent schedule
        step_counts = 0
        percent_counts = 0
        for schedule_config in obj:
            assert isinstance(schedule_config, dict)
            if "start_step" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["start_epoch", "start_update", "start_sample", "start_percent"]
                )
            elif "end_step" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["end_epoch", "end_update", "end_sample", "end_percent"]
                )
            if "start_epoch" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["start_step", "start_update", "start_sample", "start_percent"]
                )
                assert updates_per_epoch is not None
                schedule_config["start_step"] = schedule_config.pop("start_epoch") * updates_per_epoch
            elif "end_epoch" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["end_step", "end_update", "end_sample", "end_percent"]
                )
                assert updates_per_epoch is not None
                schedule_config["end_step"] = schedule_config.pop("end_epoch") * updates_per_epoch
            if "start_update" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["start_step", "start_epoch", "start_sample", "start_percent"]
                )
                schedule_config["start_step"] = schedule_config.pop("start_update")
            elif "end_update" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["end_step", "end_epoch", "end_sample", "end_percent"]
                )
                schedule_config["end_step"] = schedule_config.pop("end_update")
            if "start_epoch" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["start_step", "start_update", "start_sample", "start_percent"]
                )
                assert batch_size is not None
                schedule_config["start_step"] = int(schedule_config.pop("start_sample") / batch_size)
            elif "end_sample" in schedule_config:
                step_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["end_step", "end_epoch", "end_update", "end_percent"]
                )
                assert batch_size is not None
                schedule_config["end_step"] = int(schedule_config.pop("end_sample") / batch_size)
            elif "start_percent" in schedule_config:
                percent_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["start_step", "start_epoch", "start_update", "start_sample"]
                )
            elif "end_percent" in schedule_config:
                percent_counts += 1
                _check_mutually_exclusive_keys(
                    schedule_config=schedule_config,
                    forbidden_keys=["end_step", "end_epoch", "end_update", "end_sample"]
                )
        # if no start/end points are specified -> use step version
        if (step_counts == 0 and percent_counts == 0) or step_counts > 0:
            assert percent_counts == 0
            config_ctor = SequentialStepScheduleConfig
            ctor = SequentialStepSchedule
        elif percent_counts > 0:
            config_ctor = SequentialPercentScheduleConfig
            ctor = SequentialPercentSchedule
        else:
            raise NotImplementedError

        # sequential schedule
        schedule_configs = []
        for schedule_config in obj:
            assert "schedule" in schedule_config
            schedule = object_to_schedule(schedule_config["schedule"], **kwargs)
            cfg_kwargs = {k: v for k, v in schedule_config.items() if k != "schedule"}
            schedule_configs.append(config_ctor(schedule=schedule, **cfg_kwargs))
        return ctor(schedule_configs)

    # single schedules
    assert "kind" in obj and isinstance(obj["kind"], str)
    kind = obj.pop("kind")

    # get all names and ctors of schedules
    pascal_ctor_list = inspect.getmembers(kappaschedules.schedules, inspect.isclass)
    pascal_to_ctor = {name: ctor for name, ctor in pascal_ctor_list}
    # allow snake_case (type name is in PascalCase)
    if kind[0].islower:
        kind = kind.replace("_", "")
        snake_to_pascal = {name.lower(): name for name in pascal_to_ctor.keys()}
        assert kind in snake_to_pascal.keys(), f"invalid kind '{kind}' (possibilities: {snake_to_pascal.keys()})"
        kind = snake_to_pascal[kind]
    ctor = pascal_to_ctor[kind]

    # create SequentialScheduleConfig objects
    if ctor == SequentialPercentSchedule:
        obj["schedule_configs"] = _obj_to_schedule_configs(obj["schedule_configs"], SequentialPercentScheduleConfig)
    elif ctor == SequentialStepSchedule:
        obj["schedule_configs"] = _obj_to_schedule_configs(obj["schedule_configs"], SequentialStepScheduleConfig)

    # remove min_value/max_value if schedule doesn't need it (e.g. ConstantSchedule)
    if "max_value" in kwargs and "max_value" not in inspect.signature(ctor).parameters.keys():
        kwargs = {k: v for k, v in kwargs.items() if k != "max_value"}

    return ctor(**obj, **kwargs)

def _check_mutually_exclusive_keys(schedule_config, forbidden_keys):
    for key in schedule_config.keys():
        assert key not in forbidden_keys, f"{key} is mutually exclusive to {forbidden_keys}"

def _obj_to_schedule_configs(obj, config_ctor):
    schedule_configs = []
    for schedule_config in obj:
        schedule = object_to_schedule(schedule_config.pop("schedule"))
        schedule_configs.append(config_ctor(schedule=schedule, **schedule_config))
    return schedule_configs


from .base import IncreasingProgressSchedule
from .functional.schedules import cosine


class CosineIncreasingSchedule(IncreasingProgressSchedule):
    def _get_progress(self, step: int, total_steps: int, abs_step: int):
        return cosine(step, total_steps)



from .base import ScheduleBase


class ConstantSchedule(ScheduleBase):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __str__(self):
        return f"{type(self).__name__}(value={self.value})"

    def _get_value(self, step: int, total_steps: int, abs_step: int = None) -> float:
        return self.value



from .base import ScheduleBase


class PeriodicBoolSchedule(ScheduleBase):
    def __init__(self, initial_state, off_duration=1, on_duration=1):
        super().__init__()
        self.initial_state = initial_state
        self.off_duration = off_duration
        self.on_duration = on_duration
        self.period_duration = off_duration + on_duration

    def __str__(self):
        return (
            f"{type(self).__name__}"
            f"("
            f"intial_state={self.intial_state},"
            f"off_duration={self.off_duration},"
            f"on_duration={self.on_duration}"
            f")"
        )

    def _get_value(self, step: int, total_steps: int, abs_step: int = None) -> float:
        step_in_period = step % self.period_duration
        if self.initial_state:
            if step_in_period < self.on_duration:
                return 1.
            else:
                return 0.
        else:
            if step_in_period < self.off_duration:
                return 0.
            else:
                return 1.



from .base import ScheduleBase


class StepFixedSchedule(ScheduleBase):
    def __init__(self, start_value: float, factor: float, steps: list):
        super().__init__()
        assert isinstance(steps, list) and len(steps) > 0
        self.steps = sorted(steps)
        assert all(0. < step < 1 for step in self.steps)
        self.start_value = start_value
        self.factor = factor

    def __str__(self):
        return f"{type(self).__name__}(start_value={self.start_value}, factor={self.factor}, steps={self.steps})"

    def _get_value(self, step: int, total_steps: int, abs_step: int = None) -> float:
        progress = step / total_steps
        # search for step
        for i in range(len(self.steps)):
            if self.steps[i] > progress:
                step_idx = i
                break
        else:
            step_idx = len(self.steps)
        # round to 10th decimal place to avoid floating point precision errors
        return self.start_value * self.factor ** step_idx
